'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

exports.GomokuDirectionType = void 0;
(function (GomokuDirectionType) {
    GomokuDirectionType[GomokuDirectionType["LEFT"] = 0] = "LEFT";
    GomokuDirectionType[GomokuDirectionType["TOP_LEFT"] = 2] = "TOP_LEFT";
    GomokuDirectionType[GomokuDirectionType["TOP"] = 4] = "TOP";
    GomokuDirectionType[GomokuDirectionType["TOP_RIGHT"] = 6] = "TOP_RIGHT";
    GomokuDirectionType[GomokuDirectionType["RIGHT"] = 1] = "RIGHT";
    GomokuDirectionType[GomokuDirectionType["BOTTOM_RIGHT"] = 3] = "BOTTOM_RIGHT";
    GomokuDirectionType[GomokuDirectionType["BOTTOM"] = 5] = "BOTTOM";
    GomokuDirectionType[GomokuDirectionType["BOTTOM_LEFT"] = 7] = "BOTTOM_LEFT";
})(exports.GomokuDirectionType || (exports.GomokuDirectionType = {}));
const gomokuDirections = Array.from(Object.entries({
    [exports.GomokuDirectionType.TOP]: [-1, 0],
    [exports.GomokuDirectionType.TOP_RIGHT]: [-1, 1],
    [exports.GomokuDirectionType.RIGHT]: [0, 1],
    [exports.GomokuDirectionType.BOTTOM_RIGHT]: [1, 1],
    [exports.GomokuDirectionType.BOTTOM]: [1, 0],
    [exports.GomokuDirectionType.BOTTOM_LEFT]: [1, -1],
    [exports.GomokuDirectionType.LEFT]: [0, -1],
    [exports.GomokuDirectionType.TOP_LEFT]: [-1, -1],
}).reduce((acc, [key, value]) => {
    acc[key] = value;
    return acc;
}, []));
const leftHalfGomokuDirectionTypes = [
    exports.GomokuDirectionType.LEFT,
    exports.GomokuDirectionType.TOP_LEFT,
    exports.GomokuDirectionType.TOP,
    exports.GomokuDirectionType.TOP_RIGHT,
];
const rightHalfGomokuDirectionTypes = [
    exports.GomokuDirectionType.RIGHT,
    exports.GomokuDirectionType.BOTTOM_RIGHT,
    exports.GomokuDirectionType.BOTTOM,
    exports.GomokuDirectionType.BOTTOM_LEFT,
];
const gomokuDirectionTypes = leftHalfGomokuDirectionTypes.concat(rightHalfGomokuDirectionTypes);

class GomokuContext {
    MAX_ROW;
    MAX_COL;
    MAX_INLINE;
    MAX_NEXT_MOVER_BUFFER;
    MAX_DISTANCE_OF_NEIGHBOR;
    TOTAL_POS;
    MIDDLE_POS;
    board;
    _gomokuDirections;
    _idxMap;
    _idxMaxMoveMap;
    _placedCount;
    _nextMoverBuffer;
    constructor(MAX_ROW, MAX_COL, MAX_INLINE, MAX_NEXT_MOVER_BUFFER = 0.4, MAX_DISTANCE_OF_NEIGHBOR = 2) {
        const _TOTAL_POS = MAX_ROW * MAX_COL;
        const _MAX_NEXT_MOVER_BUFFER = Math.max(0, Math.min(1, MAX_NEXT_MOVER_BUFFER));
        const _gomokuDirections = new Int32Array(gomokuDirections.map(([dr, dc]) => dr * MAX_ROW + dc));
        const board = new Int32Array(_TOTAL_POS);
        const _idxMap = new Array(_TOTAL_POS);
        for (let r = 0; r < MAX_ROW; ++r) {
            for (let c = 0; c < MAX_COL; ++c) {
                const id = r * MAX_ROW + c;
                _idxMap[id] = [r, c];
            }
        }
        const _idxMaxMoveMap = new Array(gomokuDirectionTypes.length);
        for (const dirType of gomokuDirectionTypes) {
            const maxMoveMap = new Int32Array(_TOTAL_POS);
            const [dr, dc] = gomokuDirections[dirType];
            for (let r = 0; r < MAX_ROW; ++r) {
                for (let c = 0; c < MAX_COL; ++c) {
                    let steps = 0;
                    for (let r2 = r, c2 = c;; ++steps) {
                        r2 += dr;
                        c2 += dc;
                        if (r2 < 0 || r2 >= MAX_ROW || c2 < 0 || c2 >= MAX_COL)
                            break;
                    }
                    const id = r * MAX_ROW + c;
                    maxMoveMap[id] = steps;
                }
            }
            _idxMaxMoveMap[dirType] = maxMoveMap;
        }
        this.MAX_ROW = MAX_ROW;
        this.MAX_COL = MAX_COL;
        this.MAX_INLINE = MAX_INLINE;
        this.TOTAL_POS = _TOTAL_POS;
        this.MIDDLE_POS = _TOTAL_POS >> 1;
        this.MAX_NEXT_MOVER_BUFFER = _MAX_NEXT_MOVER_BUFFER;
        this.MAX_DISTANCE_OF_NEIGHBOR = MAX_DISTANCE_OF_NEIGHBOR;
        this.board = board;
        this._gomokuDirections = _gomokuDirections;
        this._idxMap = _idxMap;
        this._idxMaxMoveMap = _idxMaxMoveMap;
        this._placedCount = 0;
        this._nextMoverBuffer = 0;
    }
    get placedCount() {
        return this._placedCount;
    }
    get nextMoverBuffer() {
        return this._nextMoverBuffer;
    }
    reRandNextMoverBuffer() {
        this._nextMoverBuffer = Math.random() * this.MAX_NEXT_MOVER_BUFFER;
    }
    init(pieces) {
        const board = this.board;
        board.fill(-1);
        this._placedCount = pieces.length;
        for (const { r, c, p } of pieces) {
            const id = this.idx(r, c);
            board[id] = p;
        }
    }
    forward(id, p) {
        const board = this.board;
        board[id] = p;
        this._placedCount += 1;
    }
    rollback(id) {
        const board = this.board;
        board[id] = -1;
        this._placedCount -= 1;
    }
    idx(r, c) {
        return r * this.MAX_ROW + c;
    }
    idxIfValid(r, c) {
        return this.isValidPos(r, c) ? r * this.MAX_ROW + c : -1;
    }
    revIdx(id) {
        return this._idxMap[id];
    }
    isValidPos(r, c) {
        return r >= 0 && r < this.MAX_ROW && c >= 0 && c < this.MAX_COL;
    }
    isInvalidPos(r, c) {
        return r < 0 || r >= this.MAX_ROW || c < 0 || c >= this.MAX_COL;
    }
    safeMove(id, dirType, step) {
        return step <= this._idxMaxMoveMap[dirType][id]
            ? id + this._gomokuDirections[dirType] * step
            : -1;
    }
    safeMoveOneStep(id, dirType) {
        return 1 <= this._idxMaxMoveMap[dirType][id] ? id + this._gomokuDirections[dirType] : -1;
    }
    fastMove(id, dirType, step) {
        return id + this._gomokuDirections[dirType] * step;
    }
    fastMoveOneStep(id, dirType) {
        return id + this._gomokuDirections[dirType];
    }
    maxMovableSteps(id, dirType) {
        return this._idxMaxMoveMap[dirType][id];
    }
    *validNeighbors(id) {
        const { MAX_DISTANCE_OF_NEIGHBOR } = this;
        for (const dirType of gomokuDirectionTypes) {
            let id2 = id;
            for (let step = 0; step < MAX_DISTANCE_OF_NEIGHBOR; ++step) {
                id2 = this.safeMoveOneStep(id2, dirType);
                if (id2 < 0)
                    break;
                yield [id2, dirType];
            }
        }
    }
    hasPlacedNeighbors(id) {
        const { board } = this;
        for (const [id2] of this.validNeighbors(id)) {
            if (id2 >= 0 && board[id2] >= 0)
                return true;
        }
        return false;
    }
    traverseAllDirections(handle) {
        this.traverseLeftDirections(handle);
        this.traverseRightDirections(handle);
    }
    traverseLeftDirections(handle) {
        const { TOTAL_POS } = this;
        for (const dirType of leftHalfGomokuDirectionTypes) {
            for (let id = 0; id < TOTAL_POS; ++id)
                handle(id, dirType);
        }
    }
    traverseRightDirections(handle) {
        const { TOTAL_POS } = this;
        for (const dirType of rightHalfGomokuDirectionTypes) {
            for (let id = TOTAL_POS - 1; id >= 0; --id)
                handle(id, dirType);
        }
    }
}

class GomokuCountMap {
    context;
    dirCountMap;
    conShapeCountMap;
    gapShapeCountMap;
    scoreMap;
    constructor(context, scoreMap) {
        this.context = context;
        this.dirCountMap = new Array(gomokuDirections.length)
            .fill([])
            .map(() => new Uint32Array(context.TOTAL_POS));
        this.conShapeCountMap = new Array(2)
            .fill([])
            .map(() => new Array(context.MAX_INLINE + 1).fill([]).map(() => [0, 0, 0]));
        this.gapShapeCountMap = new Array(2)
            .fill([])
            .map(() => new Array(context.MAX_INLINE + 1).fill([]).map(() => [0, 0, 0]));
        this.scoreMap = scoreMap;
    }
    init() {
        this.dirCountMap.forEach(idMap => idMap.fill(0));
        this.conShapeCountMap.forEach(maps => maps.forEach(countMap => countMap.fill(0)));
        this.gapShapeCountMap.forEach(maps => maps.forEach(countMap => countMap.fill(0)));
        const { context, dirCountMap } = this;
        const { board } = context;
        context.traverseAllDirections((id, dirType) => {
            const player = board[id];
            if (player >= 0) {
                const id2 = context.safeMoveOneStep(id, dirType);
                const countMap = dirCountMap[dirType];
                countMap[id] = id2 >= 0 && board[id2] === player ? countMap[id2] + 1 : 1;
            }
        });
        context.traverseLeftDirections((id, dirType) => {
            const player = board[id];
            if (player >= 0 && dirCountMap[dirType][id] === 1) {
                const revDirType = dirType ^ 1;
                this.updateConShapeCountMap(id, revDirType, 1);
            }
        });
        context.traverseLeftDirections((id, dirType) => {
            const revDirType = dirType ^ 1;
            if (board[id] < 0) {
                const id0 = context.safeMoveOneStep(id, revDirType);
                if (id0 < 0 || board[id0] < 0)
                    return;
                const player = board[id0];
                const id2 = context.safeMoveOneStep(id, dirType);
                if (id2 < 0 || board[id2] !== player)
                    return;
                const lftCnt = dirCountMap[revDirType][id0];
                const rhtCnt = dirCountMap[dirType][id2];
                const startId = context.fastMove(id, revDirType, lftCnt);
                this.updateGapShapeCountMap(startId, lftCnt, rhtCnt, dirType, 1);
            }
        });
    }
    beforeForward(id) {
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedConShapeCountMap(id, dirType, -1);
        }
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedGapShapeCountMap(id, dirType, -1);
        }
    }
    afterForward(id) {
        const { context, dirCountMap } = this;
        const { board } = context;
        const player = board[id];
        {
            for (const dirType of gomokuDirectionTypes) {
                const id2 = context.safeMoveOneStep(id, dirType);
                const countMap = dirCountMap[dirType];
                countMap[id] = id2 >= 0 && board[id2] === player ? countMap[id2] + 1 : 1;
            }
            for (const dirType of gomokuDirectionTypes) {
                const revDirType = dirType ^ 1;
                const lftCnt = dirCountMap[revDirType][id];
                const rhtCnt = dirCountMap[dirType][id];
                const countMap = dirCountMap[dirType];
                for (let id2 = id, i = 1; i < lftCnt; ++i) {
                    id2 = context.fastMoveOneStep(id2, revDirType);
                    countMap[id2] += rhtCnt;
                }
            }
        }
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedConShapeCountMap(id, dirType, 1);
        }
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedGapShapeCountMap(id, dirType, 1);
        }
    }
    beforeRollback(id) {
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedConShapeCountMap(id, dirType, -1);
        }
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedGapShapeCountMap(id, dirType, -1);
        }
    }
    afterRollback(id) {
        const { context, dirCountMap } = this;
        {
            for (const dirType of gomokuDirectionTypes) {
                const revDirType = dirType ^ 1;
                const lftCnt = dirCountMap[revDirType][id];
                const rhtCnt = dirCountMap[dirType][id];
                const countMap = dirCountMap[dirType];
                for (let id2 = id, i = 1; i < lftCnt; ++i) {
                    id2 = context.fastMoveOneStep(id2, revDirType);
                    countMap[id2] -= rhtCnt;
                }
            }
            for (const dirType of gomokuDirectionTypes)
                dirCountMap[dirType][id] = 0;
        }
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedConShapeCountMap(id, dirType, 1);
        }
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedGapShapeCountMap(id, dirType, 1);
        }
    }
    score(scoreForPlayer) {
        const { context, conShapeCountMap, gapShapeCountMap, scoreMap } = this;
        const conCountMap = conShapeCountMap[scoreForPlayer];
        const gapCountMap = gapShapeCountMap[scoreForPlayer];
        let score = 0;
        for (let cnt = 1; cnt <= context.MAX_INLINE; ++cnt) {
            const [a, b, c] = conCountMap[cnt];
            const [x, y, z] = scoreMap.con[cnt];
            score += a * x + b * y + c * z;
        }
        for (let cnt = 1; cnt < context.MAX_INLINE; ++cnt) {
            const [a, b, c] = gapCountMap[cnt];
            const [x, y, z] = scoreMap.gap[cnt];
            score += a * x + b * y + c * z;
        }
        return score;
    }
    hasReachedTheLimit(player) {
        const con = this.conShapeCountMap[player][this.context.MAX_INLINE];
        return con[0] > 0 || con[1] > 0 || con[2] > 0;
    }
    toJSON() {
        const { dirCountMap, conShapeCountMap, gapShapeCountMap } = this;
        return { dirCountMap, conShapeCountMap, gapShapeCountMap };
    }
    updateRelatedConShapeCountMap(centerId, dirType, v) {
        const revDirType = dirType ^ 1;
        const { context, dirCountMap } = this;
        const { board } = context;
        const lftMaxMovableStep = Math.min(context.MAX_INLINE - 1, context.maxMovableSteps(centerId, revDirType));
        const rhtMaxMovableStep = Math.min(context.MAX_INLINE - 1, context.maxMovableSteps(centerId, dirType));
        const THRESHOLD = lftMaxMovableStep + rhtMaxMovableStep + 1;
        let steps = 0;
        let id = context.fastMove(centerId, revDirType, lftMaxMovableStep);
        for (; steps < THRESHOLD;) {
            if (board[id] < 0) {
                steps += 1;
                id = context.fastMoveOneStep(id, dirType);
                continue;
            }
            const rhtCnt = dirCountMap[dirType][id];
            if (steps === 0) {
                const rhtCnt = dirCountMap[dirType][id];
                const id2 = context.fastMove(id, dirType, rhtCnt - 1);
                this.updateConShapeCountMap(id2, revDirType, v);
            }
            else {
                this.updateConShapeCountMap(id, dirType, v);
            }
            id = context.fastMove(id, dirType, rhtCnt);
            steps += rhtCnt;
        }
    }
    updateConShapeCountMap(startId, dirType, v) {
        const player = this.context.board[startId];
        const cnt = this.dirCountMap[dirType][startId];
        const countOfFreeSide = this.countFreeSide(player, startId, cnt, dirType);
        const normalizedCnt = Math.min(cnt, this.context.MAX_INLINE);
        this.conShapeCountMap[player][normalizedCnt][countOfFreeSide] += v;
    }
    updateRelatedGapShapeCountMap(centerId, dirType, v) {
        const revDirType = dirType ^ 1;
        const { context, dirCountMap } = this;
        const { board } = context;
        const lftMaxMovableStep = Math.min(context.MAX_INLINE - 1, context.maxMovableSteps(centerId, revDirType));
        const rhtMaxMovableStep = Math.min(context.MAX_INLINE - 1, context.maxMovableSteps(centerId, dirType));
        const THRESHOLD = lftMaxMovableStep + rhtMaxMovableStep + 1;
        let steps = 0;
        let id = context.fastMove(centerId, revDirType, lftMaxMovableStep);
        for (; steps < THRESHOLD; steps += 1, id = context.fastMoveOneStep(id, dirType)) {
            if (board[id] < 0)
                continue;
            if (steps < 2) {
                const rhtCnt = dirCountMap[dirType][id];
                const id2 = context.fastMove(id, dirType, rhtCnt - 1);
                this.detectGapShape(id2, revDirType, v);
            }
            const lftCnt = dirCountMap[revDirType][id];
            id = context.fastMove(id, revDirType, lftCnt - 1);
            steps -= lftCnt - 1;
            break;
        }
        for (; steps < THRESHOLD;) {
            if (board[id] < 0) {
                id = context.fastMoveOneStep(id, dirType);
                steps += 1;
                continue;
            }
            this.detectGapShape(id, dirType, v);
            const rhtCnt = dirCountMap[dirType][id];
            id = context.fastMove(id, dirType, rhtCnt);
            steps += rhtCnt;
        }
    }
    detectGapShape(id, dirType, v) {
        const { context, dirCountMap } = this;
        const { board } = context;
        const lftCnt = dirCountMap[dirType][id];
        if (context.maxMovableSteps(id, dirType) < lftCnt)
            return;
        const id1 = context.fastMove(id, dirType, lftCnt);
        if (board[id1] >= 0 || context.maxMovableSteps(id1, dirType) < 1)
            return;
        const player = board[id];
        const id2 = context.fastMoveOneStep(id1, dirType);
        if (board[id2] !== player)
            return;
        const rhtCnt = dirCountMap[dirType][id2];
        this.updateGapShapeCountMap(id, lftCnt, rhtCnt, dirType, v);
    }
    updateGapShapeCountMap(startId, cnt1, cnt2, dirType, v) {
        const threshold = this.context.MAX_INLINE - 1;
        if (cnt1 >= threshold || cnt2 >= threshold)
            return;
        const player = this.context.board[startId];
        const cnt = cnt1 + cnt2;
        const countOfFreeSide = this.countFreeSide(player, startId, cnt + 1, dirType);
        const normalizedCnt = Math.min(cnt, this.context.MAX_INLINE);
        this.gapShapeCountMap[player][normalizedCnt][countOfFreeSide] += v;
    }
    countFreeSide(player, startId, cnt, dirType) {
        const { context } = this;
        const { board, MAX_INLINE } = context;
        const revDirType = dirType ^ 1;
        const id0 = context.safeMoveOneStep(startId, revDirType);
        const isFreeSide0 = id0 >= 0 && board[id0] < 0 ? 1 : 0;
        const id2 = context.safeMove(startId, dirType, cnt);
        const isFreeSide2 = id2 >= 0 && board[id2] < 0 ? 1 : 0;
        const countOfFreeSide = isFreeSide0 + isFreeSide2;
        if (countOfFreeSide === 0)
            return 0;
        let maxPossibleCnt = cnt + countOfFreeSide;
        if (maxPossibleCnt < MAX_INLINE) {
            const revPlayer = player ^ 1;
            if (isFreeSide0) {
                const maxMovableSteps = context.maxMovableSteps(id0, revDirType);
                for (let id = id0, step = 0; maxPossibleCnt < MAX_INLINE && step < maxMovableSteps; ++step, ++maxPossibleCnt) {
                    id = context.fastMoveOneStep(id, revDirType);
                    if (board[id] === revPlayer)
                        break;
                }
            }
            if (isFreeSide2) {
                const maxMovableSteps = context.maxMovableSteps(id2, dirType);
                for (let id = id2, step = 0; maxPossibleCnt < MAX_INLINE && step < maxMovableSteps; ++step, ++maxPossibleCnt) {
                    id = context.fastMoveOneStep(id, dirType);
                    if (board[id] === revPlayer)
                        break;
                }
            }
        }
        return maxPossibleCnt < MAX_INLINE ? 0 : countOfFreeSide;
    }
}

class GomokuState {
    context;
    countMap;
    candidateMap;
    constructor(context, scoreMap) {
        this.context = context;
        this.countMap = new GomokuCountMap(context, scoreMap);
        this.candidateMap = new Map();
    }
    init(pieces = []) {
        const { context, countMap, candidateMap } = this;
        context.init(pieces);
        countMap.init();
        candidateMap.clear();
        const stateScore0 = this.countMap.score(0);
        const stateScore1 = this.countMap.score(1);
        for (const { r, c } of pieces) {
            const id = context.idx(r, c);
            for (const [id2] of context.validNeighbors(id)) {
                if (context.board[id2] < 0 && !candidateMap.has(id2)) {
                    const score = this._scoreForCandidate(stateScore0, stateScore1, id2);
                    candidateMap.set(id2, score);
                }
            }
        }
    }
    forward(id, player) {
        const { context, candidateMap } = this;
        if (id < 0 || context.board[id] >= 0)
            return;
        this._forward(id, player);
        candidateMap.delete(id);
        for (const [id2] of context.validNeighbors(id)) {
            if (context.board[id2] < 0 && !candidateMap.has(id2))
                candidateMap.set(id2, [0, 0, 0, 0]);
        }
        this._reAppriseCandidates(id);
    }
    rollback(id) {
        const { context, candidateMap } = this;
        if (id < 0 || context.board[id] < 0)
            return;
        this._rollback(id);
        if (context.hasPlacedNeighbors(id))
            candidateMap.set(id, [0, 0, 0, 0]);
        for (const [id2] of context.validNeighbors(id)) {
            if (context.board[id2] >= 0 || !context.hasPlacedNeighbors(id2)) {
                candidateMap.delete(id2);
            }
        }
        this._reAppriseCandidates(id);
    }
    expand(nextPlayer, scoreForPlayer) {
        const { context, countMap, candidateMap } = this;
        const stateScore0 = countMap.score(0);
        const stateScore1 = countMap.score(1);
        if (context.board[context.MIDDLE_POS] < 0 && !candidateMap.has(context.MIDDLE_POS)) {
            candidateMap.set(context.MIDDLE_POS, this._scoreForCandidate(stateScore0, stateScore1, context.MIDDLE_POS));
        }
        const player0 = nextPlayer << 1;
        const player1 = player0 ^ 1;
        const nextMoverFac = 1 + context.nextMoverBuffer;
        const candidates = [];
        for (const [id, candidateScore] of candidateMap.entries()) {
            if (context.board[id] >= 0)
                continue;
            const scores = [
                stateScore0 + candidateScore[player0],
                stateScore1 + candidateScore[player1],
            ];
            const score0 = scores[scoreForPlayer];
            const score1 = scores[scoreForPlayer ^ 1];
            const score = nextPlayer === scoreForPlayer
                ? score0 - score1 * nextMoverFac
                : score0 * nextMoverFac - score1;
            candidates.push({ id, score });
        }
        return candidates;
    }
    score(currentPlayer, scoreForPlayer) {
        const score0 = this.countMap.score(scoreForPlayer);
        const score1 = this.countMap.score(scoreForPlayer ^ 1);
        const nextMoverFac = 1 + this.context.nextMoverBuffer;
        return currentPlayer === scoreForPlayer
            ? score0 - score1 * nextMoverFac
            : score0 * nextMoverFac - score1;
    }
    isWin(currentPlayer) {
        return this.countMap.hasReachedTheLimit(currentPlayer);
    }
    isDraw() {
        return this.context.placedCount === this.context.TOTAL_POS;
    }
    isFinal() {
        const { context, countMap } = this;
        if (context.placedCount === context.TOTAL_POS)
            return true;
        if (countMap.hasReachedTheLimit(0) || countMap.hasReachedTheLimit(1))
            return true;
        return false;
    }
    _scoreForCandidate(stateScore0, stateScore1, id) {
        const result = [0, 0, 0, 0];
        const { context, countMap } = this;
        for (let p = 0; p < 2; ++p) {
            this._beforeForward(id);
            context.forward(id, p);
            this._afterForward(id);
            result[p << 1] = countMap.score(0) - stateScore0;
            result[(p << 1) | 1] = countMap.score(1) - stateScore1;
            this._beforeRollback(id);
            context.rollback(id);
            this._afterRollback(id);
        }
        return result;
    }
    _reAppriseCandidates(centerId) {
        const { context, countMap, candidateMap } = this;
        const stateScore0 = countMap.score(0);
        const stateScore1 = countMap.score(1);
        if (this.candidateMap.has(centerId)) {
            const score = this._scoreForCandidate(stateScore0, stateScore1, centerId);
            candidateMap.set(centerId, score);
        }
        const THRESHOLD = context.MAX_INLINE * 2 - 3;
        for (const dirType of gomokuDirectionTypes) {
            let id = centerId;
            const maxStep = Math.min(THRESHOLD, context.maxMovableSteps(id, dirType));
            for (let step = 0; step < maxStep; ++step) {
                id = context.fastMoveOneStep(id, dirType);
                if (candidateMap.has(id)) {
                    const score = this._scoreForCandidate(stateScore0, stateScore1, id);
                    candidateMap.set(id, score);
                }
            }
        }
    }
    _forward(id, player) {
        this._beforeForward(id);
        this.context.forward(id, player);
        this._afterForward(id);
    }
    _rollback(id) {
        this._beforeRollback(id);
        this.context.rollback(id);
        this._afterRollback(id);
    }
    _beforeForward(id) {
        this.countMap.beforeForward(id);
    }
    _afterForward(id) {
        this.countMap.afterForward(id);
    }
    _beforeRollback(id) {
        this.countMap.beforeRollback(id);
    }
    _afterRollback(id) {
        this.countMap.afterRollback(id);
    }
}

const createScoreMap = (TOTAL_POS, MAX_INLINE) => {
    const gap = new Array(MAX_INLINE + 1).fill([]).map(() => [0, 0, 0]);
    const con = new Array(MAX_INLINE + 1).fill([]).map(() => [0, 0, 0]);
    let baseValue = 2;
    for (let cnt = 1; cnt < MAX_INLINE; ++cnt) {
        gap[cnt] = [0, baseValue / 2, baseValue];
        con[cnt] = [0, baseValue, baseValue * 2];
        const scale = Math.max(4, Math.round(TOTAL_POS / (cnt * cnt)));
        baseValue *= scale;
    }
    const _v = con[MAX_INLINE - 1][1];
    gap[MAX_INLINE - 1] = [_v, _v, _v];
    gap[MAX_INLINE] = [_v, _v, _v];
    con[MAX_INLINE] = [baseValue, baseValue, baseValue];
    return { con: con, gap };
};

class GomokuSolution {
    MAX_DEPTH;
    context;
    state;
    stateCache;
    scoreForPlayer;
    bestMoveId;
    constructor(MAX_ROW, MAX_COL, MAX_INLINE = 5, MAX_DEPTH = 3, MAX_NEXT_MOVER_BUFFER, scoreMap) {
        const context = new GomokuContext(MAX_ROW, MAX_COL, MAX_INLINE, MAX_NEXT_MOVER_BUFFER);
        const _scoreMap = scoreMap ?? createScoreMap(context.TOTAL_POS, context.MAX_INLINE);
        this.MAX_DEPTH = Math.max(1, Math.round(MAX_DEPTH));
        this.context = context;
        this.state = new GomokuState(context, _scoreMap);
        this.stateCache = new Map();
        this.scoreForPlayer = -1;
        this.bestMoveId = null;
    }
    init(pieces) {
        this.state.init(pieces);
        this.stateCache.clear();
    }
    forward(r, c, p) {
        const id = this.context.idx(r, c);
        this.state.forward(id, p);
    }
    rollback(r, c) {
        const id = this.context.idx(r, c);
        this.state.rollback(id);
    }
    minimaxSearch(nextPlayer) {
        if (this.state.isFinal())
            return [-1, -1];
        this.stateCache.clear();
        this.scoreForPlayer = nextPlayer;
        this.bestMoveId = null;
        this.context.reRandNextMoverBuffer();
        this.alphaBeta(nextPlayer, Number.NEGATIVE_INFINITY, Number.POSITIVE_INFINITY, 0, 0);
        if (!this.bestMoveId)
            return [-1, -1];
        const [r, c] = this.context.revIdx(this.bestMoveId);
        return [r, c];
    }
    alphaBeta(player, alpha, beta, cur, stateScore) {
        const { state, scoreForPlayer } = this;
        if (state.isWin(this.scoreForPlayer))
            return Number.POSITIVE_INFINITY;
        if (state.isWin(this.scoreForPlayer ^ 1))
            return Number.NEGATIVE_INFINITY;
        if (cur === this.MAX_DEPTH || state.isDraw())
            return stateScore;
        const candidates = state.expand(player, scoreForPlayer);
        if (player === scoreForPlayer) {
            candidates.sort((x, y) => y.score - x.score);
        }
        else {
            candidates.sort((x, y) => x.score - y.score);
        }
        if (cur === 0)
            this.bestMoveId = candidates[0].id;
        for (const { id, score } of candidates) {
            state.forward(id, player);
            const gamma = this.alphaBeta(player ^ 1, alpha, beta, cur + 1, score);
            state.rollback(id);
            if (player === scoreForPlayer) {
                if (alpha < gamma) {
                    alpha = gamma;
                    if (cur === 0)
                        this.bestMoveId = id;
                }
            }
            else {
                if (beta > gamma) {
                    beta = gamma;
                }
            }
            if (beta <= alpha)
                break;
        }
        return player === scoreForPlayer ? alpha : beta;
    }
}

class GomokuStateCompressor {
    MOVE_STEP_BIT_BASE;
    MOVE_STEP_MASK;
    MOVE_STEP_MINIMUM_FLAG;
    INITIAL_STATE = 1n;
    constructor(TOTAL_POS) {
        let COMPRESS_MOVE_BIT_BASE = 1n;
        while (1n << COMPRESS_MOVE_BIT_BASE < TOTAL_POS)
            COMPRESS_MOVE_BIT_BASE += 1n;
        this.MOVE_STEP_BIT_BASE = COMPRESS_MOVE_BIT_BASE;
        this.MOVE_STEP_MASK = (1n << COMPRESS_MOVE_BIT_BASE) - 1n;
        this.MOVE_STEP_MINIMUM_FLAG = 1n << COMPRESS_MOVE_BIT_BASE;
    }
    compress(countInState, state, newMoveStepId) {
        if (countInState === 0)
            return newMoveStepId + this.MOVE_STEP_MINIMUM_FLAG;
        if (countInState === 1)
            return (state << this.MOVE_STEP_BIT_BASE) | newMoveStepId;
        const rhtCnt = countInState >> 1;
        const RIGHT_BITS = this.MOVE_STEP_BIT_BASE * BigInt(rhtCnt);
        const rht = state & ((1n << RIGHT_BITS) - 1n);
        if (countInState & 1) {
            const rhtMoveIds = this.decompressMoveIds(rhtCnt, rht, newMoveStepId);
            const newRht = this.compressMoveIds(rhtMoveIds);
            return ((state ^ rht) << this.MOVE_STEP_BIT_BASE) | newRht;
        }
        else {
            const lftCnt = countInState - rhtCnt;
            const lft = state >> RIGHT_BITS;
            const lftMoveIds = this.decompressMoveIds(lftCnt, lft, newMoveStepId);
            const NEW_LEFT_BITS = this.MOVE_STEP_BIT_BASE * BigInt(lftCnt + 1);
            const newLft = this.compressMoveIds(lftMoveIds) | (1n << NEW_LEFT_BITS);
            return (newLft << RIGHT_BITS) | rht;
        }
    }
    decompressMoveIds(cnt, state, newMoveStepId) {
        const { MOVE_STEP_BIT_BASE, MOVE_STEP_MASK } = this;
        const moveIds = [];
        let i = 0;
        let v = state;
        let id = 0n;
        for (; i < cnt; ++i) {
            id = v & MOVE_STEP_MASK;
            v >>= MOVE_STEP_BIT_BASE;
            if (id < newMoveStepId)
                break;
            moveIds.push(id);
        }
        moveIds.push(newMoveStepId);
        if (i < cnt)
            moveIds.push(id);
        for (i += 1; i < cnt; ++i) {
            id = v & MOVE_STEP_MASK;
            v >>= MOVE_STEP_BIT_BASE;
            moveIds.push(id);
        }
        return moveIds.reverse();
    }
    compressMoveIds(moveIds) {
        const { MOVE_STEP_BIT_BASE } = this;
        let result = 0n;
        for (const id of moveIds) {
            result = (result << MOVE_STEP_BIT_BASE) | id;
        }
        return result;
    }
}

exports.GomokuContext = GomokuContext;
exports.GomokuCountMap = GomokuCountMap;
exports.GomokuSolution = GomokuSolution;
exports.GomokuState = GomokuState;
exports.GomokuStateCompressor = GomokuStateCompressor;
exports.createScoreMap = createScoreMap;
exports.gomokuDirectionTypes = gomokuDirectionTypes;
exports.gomokuDirections = gomokuDirections;
exports.leftHalfGomokuDirectionTypes = leftHalfGomokuDirectionTypes;
exports.rightHalfGomokuDirectionTypes = rightHalfGomokuDirectionTypes;
